#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>          
#include <sys/stat.h>      
#include <semaphore.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>

#define MAXITEMS 3

typedef struct {
    int is_null;
    char value;
} itemtype;

sem_t *empty;
sem_t *full;
sem_t *mutex;

int ends[2];

void produce(itemtype *item)
{
    static int i = 0;
    const char *text ="AHOOOJ!";

    if(i != strlen(text)) {
        item->is_null = 0;
        item->value = text[i];
    }
    else {
        item->is_null = 1;
        item->value = '\0';
    }
    i++;
}

void enter_item(itemtype item)
{
    write(ends[1],&item, sizeof(itemtype));
    printf("Item Entered: { %i, %c }\n",item.is_null,item.value);
}

void remove_item(itemtype *item)
{
    read(ends[0],item, sizeof(itemtype));
}

int consume_item(itemtype item)
{
    printf("Item Consumed: { %i, %c }\n",item.is_null,item.value);
    sleep(2);
    if(item.is_null)
        return 1;
    return 0;
}

void producer(void)
{
    itemtype item;
    while(1) {
        produce(&item);
        sem_wait(empty);
        sem_wait(mutex); 
        enter_item(item);
        sem_post(mutex);
        sem_post(full);
        if(item.is_null)
           break;
    }
}

void consumer(void) 
{
    itemtype item;
    while (1) {
        sem_wait(full);
        sem_wait(mutex);
        remove_item(&item);
        sem_post(mutex);
        sem_post(empty);
        if(consume_item(item))
            break;
    }
}

int main(void)
{
    pid_t pid;
    int status;
    empty = sem_open("empty", O_CREAT, 0, MAXITEMS);
    if (empty == SEM_FAILED) {
        perror("Error creating semaphore empty:");
        exit(1);
    }
    full = sem_open("full",O_CREAT,0,0);
    if (full == SEM_FAILED) {
        perror("Error creating semaphore full:");
        exit(1);
    }
    mutex = sem_open("mutex",O_CREAT,0,1);
    if (mutex == SEM_FAILED) {
        perror("Error creating semaphore mutex:");
        exit(1);
    }

    if (pipe(ends) < 0) {
            perror("Pipe creation failed");
            exit(1);
    }

    if(!(pid = fork()))
    {
        close(ends[0]);
        producer();
        close(ends[1]);
        printf("producer end\n");
    }
    else
    {
        printf("pid = %i\n",pid);
        close(ends[1]);
        consumer();
        close(ends[0]);
        printf("consumer end\n");
        waitpid(pid,&status,0);
        sem_close(empty);
        sem_close(full);
        sem_close(mutex);
    }
    return 0;
}
